from collections.abc import Iterable

import numpy as np


class EvolutionResult:
    """Temporal distribution of various disk parameters

    Parameters
    ----------
    states: Iterable of states, e.g. list of Freddi
        States to be stored as results

    Methods
    -------
    flux(lmbd, region, phase) : array
        Optical flux of the disk

    """
    def __init__(self, states):
        self.__states = tuple(states)
        self.__len_states = len(self.__states)

    def flux(self, lmbd, region='hot', phase=None) -> np.ndarray:
        """Optical flux of the disk

        Parameters
        ----------
        lmbd : array_like
            Observation wavelength
        region: str, optional
            Object optically luminous component, could be one of: `hot`, `cold`,
            `disk`, `star`, `all`
        phase: array-like or float or None, optional
            Phase of the observation in radians, must be specified if `region`
            is `star` or `all`

        Returns
        -------
        array

        """
        if isinstance(phase, Iterable):
            phase = np.asarray(phase)
            if phase.shape != (self.__len_states, ):
                raise ValueError('phase length should be {}'.format(self.__len_states))
        else:
            phase = [phase] * self.__len_states
        lmbd = np.asarray(lmbd)
        arr = np.empty((self.__len_states,) + lmbd.shape, dtype=np.float)
        for i in range(self.__len_states):
            arr[i] = self.__states[i].flux(lmbd, region, phase[i])
        return arr

    def __getattr__(self, attr) -> np.ndarray:
        first_val = np.asarray(getattr(self.__states[0], attr))
        if first_val.ndim == 0:
            arr = np.empty(self.__len_states, dtype=first_val.dtype)
            arr[0] = first_val
            for i in range(1, self.__len_states):
                arr[i] = getattr(self.__states[i], attr)
            return arr
        elif first_val.ndim == 1:
            arr = np.full((self.__len_states, self.__states[0].Nx), np.nan, dtype=first_val.dtype)
            for i, state in enumerate(self.__states):
                value = np.asarray(getattr(state, attr))
                idx = slice(state.first, state.last + 1)
                arr[i, idx] = value[idx]
            return arr
        else:
            raise ValueError("{}.ndim > 1 ({})".format(attr, first_val.ndim))



from functools import partial

import numpy as np

from ._freddi import _Freddi, _FreddiNeutronStar
from .evolution_result import EvolutionResult


class _MetaFreddi(type(_Freddi)):
    def __new__(mcs, name, bases, attrs, boost_cls=object):
        attrs['_boost_class'] = boost_cls
        test_obj = boost_cls(**boost_cls._required_args())
        for attr_name in dir(boost_cls):
            if attr_name.startswith('_'):
                continue
            cls_attr = getattr(boost_cls, attr_name)
            if not isinstance(cls_attr, property):
                continue
            attrs[attr_name] = property(partial(mcs._boost_cls_property, attr_name))
            obj_attr = getattr(test_obj, attr_name)
            if isinstance(obj_attr, np.ndarray):
                attrs['first_' + attr_name] = property(partial(mcs._first_last_getter, 'first', attr_name))
                attrs['last_' + attr_name] = property(partial(mcs._first_last_getter, 'last', attr_name))

        return super().__new__(mcs, name, bases, attrs)

    def __init__(cls, name, bases, attrs, boost_cls=object):
        super().__init__(name, bases, attrs)

    @staticmethod
    def _boost_cls_property(attr_name, self):
        return getattr(self._freddi, attr_name)

    @staticmethod
    def _first_last_getter(first_or_last, attr_name, self):
        return getattr(self._freddi, attr_name)[getattr(self._freddi, first_or_last)]


class _BasePyFreddi:
    def __init__(self, **kwargs):
        self._freddi = self._boost_class(**kwargs)
        self._kwargs = self._freddi._kwargs

    @classmethod
    def _from_boost(cls, boost_freddi):
        obj = cls.__new__(cls)
        obj._freddi = boost_freddi
        return obj

    @classmethod
    def alt(cls, **kwargs):
        """Alternative Freddi constructor accepting astropy Quantity and str

        All physical values can be passed as `astropy.units.Quantity`, and all
        string values can be passed as `str`. Needs `astropy` module to run

        Parameters
        ----------
        **kwargs :
            `Freddi` arguments

        Returns
        -------
        Freddi

        """
        from astropy.units import Quantity

        for item, value in kwargs.items():
            if isinstance(value, Quantity):
                kwargs[item] = value.cgs.value
                continue
            if isinstance(value, str):
                kwargs[item] = value.encode()
        return cls(**kwargs)

    def evolve(self):
        """Calculate disk evolution

        Returns
        -------
        EvolutionResults

        """
        return EvolutionResult(self)

    def _flux_hot(self, lmbd, phase):
        del phase
        return self._freddi._flux_hot(lmbd)

    def _flux_cold(self, lmbd, phase):
        del phase
        return self._freddi._flux_cold(lmbd)

    def _flux_star(self, lmbd, phase):
        return self._freddi._flux_star(lmbd, phase)

    def flux(self, lmbd, region='hot', phase=None):
        region = region.lower()
        if 'hot'.startswith(region):
            flux = self._flux_hot
        elif 'cold'.startswith(region):
            flux = self._flux_cold
        elif 'disk'.startswith(region):
            def flux(lmbd, phase):
                return self._flux_hot(lmbd, phase) + self._flux_cold(lmbd, phase)
        elif 'star'.startswith(region):
            if phase is None:
                raise ValueError('Phase must be specified if star flux is required')
            flux = self._flux_star
        elif 'all'.startswith(region):
            def flux(lmbd, phase):
                return self._flux_hot(lmbd, phase) + self._flux_cold(lmbd, phase) + self._flux_star(lmbd, phase)
        else:
            raise ValueError(f'Zone {region} is not supported')

        lmbd = np.asarray(lmbd)
        arr = np.empty_like(lmbd, dtype=float)
        for i, x in np.ndenumerate(lmbd):
            arr[i] = flux(float(x), phase)
        return arr

    def __iter__(self):
        for value in self._freddi:
            value = self._from_boost(value)
            yield value


class Freddi(_BasePyFreddi, metaclass=_MetaFreddi, boost_cls=_Freddi):
    pass


class FreddiNeutronStar(_BasePyFreddi, metaclass=_MetaFreddi, boost_cls=_FreddiNeutronStar):
    pass



#!/usr/bin/env python

import os

from skbuild import setup
from setuptools.command.build_ext import build_ext


with open('Readme.md') as fh:
    readme = fh.read()


setup(
    name='freddi',
    version='2.0.0b1',
    url='http://xray.sai.msu.ru/~malanchev/freddi/',
    license='MIT',
    author='Konstantin Malanchev',
    author_email='malanchev@sai.msu.ru',
    description='Compute FRED light curves of LMXBs outbursts',
    long_description=readme,
    long_description_content_type='text/markdown',
    package_dir={'': 'python'},
    packages=['freddi'],
    install_requires=['numpy'],
    setup_requires=['numpy'],
    tests_require=['parameterized'],
    python_requires='>=3.5',
    test_suite='test',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    keywords='science astrophysics accretion',
)

